/*
 *  Copyright (c) 2013 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 *
 */

#include "dl/sp/api/omxSP.h"

// For ArmCpuCaps()
#include <stdio.h>
#include <string.h>

// based on libvpx arm_cpudetect.c
static int ArmCpuCaps(const char* cpuinfo_name) {
  char cpuinfo_line[512];
  FILE* f = fopen(cpuinfo_name, "r");
  if (!f) {
    // Assume Neon if /proc/cpuinfo is unavailable.
    // This will occur for Chrome sandbox for Pepper or Render process.
    return 1;
  }
  while (fgets(cpuinfo_line, sizeof(cpuinfo_line) - 1, f)) {
    if (memcmp(cpuinfo_line, "Features", 8) == 0) {
      char* p = strstr(cpuinfo_line, " neon");
      if (p && (p[5] == ' ' || p[5] == '\n')) {
        fclose(f);
        return 1;
      }
      // aarch64 uses asimd for Neon.
      p = strstr(cpuinfo_line, " asimd");
      if (p && (p[6] == ' ' || p[6] == '\n')) {
        fclose(f);
        return 1;
      }
    }
  }
  fclose(f);
  return 0;
}

int omxSP_HasArmNeon() {
#if defined(__arm__) || defined(__aarch64__)
// gcc -mfpu=neon defines __ARM_NEON__
// __ARM_NEON__ generates code that requires Neon.  NaCL also requires Neon.
// For Linux, /proc/cpuinfo can be tested but without that assume Neon.
#if defined(__ARM_NEON__) || defined(__native_client__) || !defined(__linux__)
  return 1;
// For aarch64(arm64), /proc/cpuinfo's feature is not complete, e.g. no neon
// flag in it.
// So for aarch64, neon enabling is hard coded here.
#elif defined(__aarch64__)
  return 1;
#else
  // Linux arm parse text file for neon detect.
  return ArmCpuCaps("/proc/cpuinfo");
#endif
#endif  // __arm__
}

static void SetFFTRoutines() {
  /*
   * Choose the correct (NEON or non-NEON) routines for both the
   * forward and inverse FFTs
   */
  if (omxSP_HasArmNeon()) {
    omxSP_FFTFwd_RToCCS_F32 = omxSP_FFTFwd_RToCCS_F32_Sfs;
    omxSP_FFTInv_CCSToR_F32 = omxSP_FFTInv_CCSToR_F32_Sfs;
  } else {
    omxSP_FFTFwd_RToCCS_F32 = omxSP_FFTFwd_RToCCS_F32_Sfs_vfp;
    omxSP_FFTInv_CCSToR_F32 = omxSP_FFTInv_CCSToR_F32_Sfs_vfp;
  }
}

/*
 * FIXME: It would be beneficial to use the GCC ifunc attribute to
 * select the appropriate function at load time. This is apparently
 * not supported on Android at this time. (Compiler warning that the
 * ifunc attribute is ignored.)
 */

/*
 * Forward FFT.  Detect if NEON is supported and update function
 * pointers to the correct routines for both the forward and inverse
 * FFTs.  Then run the forward FFT routine.
 */
static OMXResult DetectForwardRealFFT(
    const OMX_F32* pSrc,
    OMX_F32* pDst,
    const OMXFFTSpec_R_F32* pFFTSpec) {
  SetFFTRoutines();
  return omxSP_FFTFwd_RToCCS_F32(pSrc, pDst, pFFTSpec);
}

/*
 * Inverse FFT.  Detect if NEON is supported and update function
 * pointers to the correct routines for both the forward and inverse
 * FFTs.  Then run the inverse FFT routine.
 */
static OMXResult DetectInverseRealFFT(
    const OMX_F32* pSrc,
    OMX_F32* pDst,
    const OMXFFTSpec_R_F32* pFFTSpec) {
  SetFFTRoutines();
  return omxSP_FFTInv_CCSToR_F32(pSrc, pDst, pFFTSpec);
}

/*
 * Implementation of the forward and inverse real float FFT.
 * Initialize to detection routine which will update the pointer to
 * the correct routine and then call the correct one.
 */
OMXResult (*omxSP_FFTFwd_RToCCS_F32)(
    const OMX_F32* pSrc,
    OMX_F32* pDst,
    const OMXFFTSpec_R_F32* pFFTSpec) = DetectForwardRealFFT;

OMXResult (*omxSP_FFTInv_CCSToR_F32)(
    const OMX_F32* pSrc,
    OMX_F32* pDst,
    const OMXFFTSpec_R_F32* pFFTSpec) = DetectInverseRealFFT;
