/*
 * Copyright 2016 Google Inc.
 *
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

#include "SkCpu.h"
#include "SkOnce.h"

#if defined(SK_CPU_X86)
    #if defined(SK_BUILD_FOR_WIN32)
        #include <intrin.h>
        static void cpuid (uint32_t abcd[4]) { __cpuid  ((int*)abcd, 1);    }
        static void cpuid7(uint32_t abcd[4]) { __cpuidex((int*)abcd, 7, 0); }
        static uint64_t xgetbv(uint32_t xcr) { return _xgetbv(xcr); }
    #else
        #include <cpuid.h>
        #if !defined(__cpuid_count)  // Old Mac Clang doesn't have this defined.
            #define  __cpuid_count(eax, ecx, a, b, c, d) \
                __asm__("cpuid" : "=a"(a), "=b"(b), "=c"(c), "=d"(d) : "0"(eax), "2"(ecx))
        #endif
        static void cpuid (uint32_t abcd[4]) { __get_cpuid(1, abcd+0, abcd+1, abcd+2, abcd+3); }
        static void cpuid7(uint32_t abcd[4]) {
            __cpuid_count(7, 0, abcd[0], abcd[1], abcd[2], abcd[3]);
        }
        static uint64_t xgetbv(uint32_t xcr) {
            uint32_t eax, edx;
            __asm__ __volatile__ ( "xgetbv" : "=a"(eax), "=d"(edx) : "c"(xcr));
            return (uint64_t)(edx) << 32 | eax;
        }
    #endif

    static uint32_t read_cpu_features() {
        uint32_t features = 0;
        uint32_t abcd[4] = {0,0,0,0};

        // You might want to refer to http://www.sandpile.org/x86/cpuid.htm

        cpuid(abcd);
        if (abcd[3] & (1<<25)) { features |= SkCpu:: SSE1; }
        if (abcd[3] & (1<<26)) { features |= SkCpu:: SSE2; }
        if (abcd[2] & (1<< 0)) { features |= SkCpu:: SSE3; }
        if (abcd[2] & (1<< 9)) { features |= SkCpu::SSSE3; }
        if (abcd[2] & (1<<19)) { features |= SkCpu::SSE41; }
        if (abcd[2] & (1<<20)) { features |= SkCpu::SSE42; }

        if ((abcd[2] & (3<<26)) == (3<<26) && (xgetbv(0) & 6) == 6) {  // XSAVE + OSXSAVE
            if (abcd[2] & (1<<28)) { features |= SkCpu:: AVX; }
            if (abcd[2] & (1<<29)) { features |= SkCpu::F16C; }
            if (abcd[2] & (1<<12)) { features |= SkCpu:: FMA; }

            cpuid7(abcd);
            if (abcd[1] & (1<<5)) { features |= SkCpu::AVX2; }
            if (abcd[1] & (1<<3)) { features |= SkCpu::BMI1; }
            if (abcd[1] & (1<<8)) { features |= SkCpu::BMI2; }
        }
        return features;
    }

#elif defined(SK_CPU_ARM32)         && \
      defined(SK_BUILD_FOR_ANDROID) && \
     !defined(SK_BUILD_FOR_ANDROID_FRAMEWORK)
    #include "cpu-features.h"

    static uint32_t read_cpu_features() {
        uint32_t features = 0;

        uint64_t android_features = android_getCpuFeatures();
        if (android_features & ANDROID_CPU_ARM_FEATURE_NEON    ) { features |= SkCpu::NEON    ; }
        if (android_features & ANDROID_CPU_ARM_FEATURE_NEON_FMA) { features |= SkCpu::NEON_FMA; }
        if (android_features & ANDROID_CPU_ARM_FEATURE_VFP_FP16) { features |= SkCpu::VFP_FP16; }
        return features;
    }

#elif defined(SK_CPU_ARM32)         && \
      !defined(SK_BUILD_FOR_ANDROID)
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

    static uint32_t read_cpu_features() {
        uint32_t features = 0;

        // If we fail any of the following, assume we don't have NEON/VFPv4 instructions
        // This allows us to return immediately in case of error.
        bool have_neon = false;
        bool have_vfpv4 = false;

        // There is no user-accessible CPUID instruction on ARM that we can use.
        // Instead, we must parse /proc/cpuinfo and look for the 'neon' feature.
        // For example, here's a typical output (Nexus S running ICS 4.0.3):
        /*
        Processor       : ARMv7 Processor rev 2 (v7l)
        BogoMIPS        : 994.65
        Features        : swp half thumb fastmult vfp edsp thumbee neon vfpv3
        CPU implementer : 0x41
        CPU architecture: 7
        CPU variant     : 0x2
        CPU part        : 0xc08
        CPU revision    : 2

        Hardware        : herring
        Revision        : 000b
        Serial          : 3833c77d6dc000ec
        */
        char   buffer[4096];

        do {
            // open /proc/cpuinfo
            int fd = TEMP_FAILURE_RETRY(open("/proc/cpuinfo", O_RDONLY));
            if (fd < 0) {
                SkDebugf("Could not open /proc/cpuinfo: %s\n", strerror(errno));
                break;
            }

            // Read the file. To simplify our search, we're going to place two
            // sentinel '\n' characters: one at the start of the buffer, and one at
            // the end. This means we reserve the first and last buffer bytes.
            buffer[0] = '\n';
            int size = TEMP_FAILURE_RETRY(read(fd, buffer+1, sizeof(buffer)-2));
            close(fd);

            if (size < 0) {  // should not happen
                SkDebugf("Could not read /proc/cpuinfo: %s\n", strerror(errno));
                break;
            }

            SkDebugf("START /proc/cpuinfo:\n%.*s\nEND /proc/cpuinfo\n",
                    size, buffer+1);

            // Compute buffer limit, and place final sentinel
            char* buffer_end = buffer + 1 + size;
            buffer_end[0] = '\n';

            // Now, find a line that starts with "Features", i.e. look for
            // '\nFeatures ' in our buffer.
            const char features[] = "\nFeatures\t";
            const size_t features_len = sizeof(features)-1;

            char*  line = (char*) memmem(buffer, buffer_end - buffer,
                                        features, features_len);
            if (line == nullptr) {  // Weird, no Features line, bad kernel?
                SkDebugf("Could not find a line starting with 'Features'"
                "in /proc/cpuinfo ?\n");
                break;
            }

            line += features_len;  // Skip the "\nFeatures\t" prefix

            // Find the end of the current line
            char* line_end = (char*) memchr(line, '\n', buffer_end - line);
            if (line_end == nullptr)
                line_end = buffer_end;

            // Now find an instance of 'neon' in the flags list. We want to
            // ensure it's only 'neon' and not something fancy like 'noneon'
            // so check that it follows a space.
            const char neon[] = " neon";
            const size_t neon_len = sizeof(neon)-1;
            const char* flag = (const char*) memmem(line, line_end - line,
                                                    neon, neon_len);
            // Ensure it is followed by a space or a newline.
            if (flag != nullptr
                && (flag[neon_len] == ' ' || flag[neon_len] == '\n')) {
                // Fine, we support Arm NEON !
                have_neon = true;
            }

            // Now find an instance of 'vfpv4' in the flags list. We want to
            // ensure it's only 'vfpv4' and not something fancy like 'novfpv4'
            // so check that it follows a space.
            const char vfpv4[] = " vfpv4";
            const size_t vfpv4_len = sizeof(vfpv4)-1;
            const char* vflag = (const char*) memmem(line, line_end - line,
                                                     vfpv4, vfpv4_len);
            // Ensure it is followed by a space or a newline.
            if (vflag != nullptr
                && (vflag[vfpv4_len] == ' ' || vflag[vfpv4_len] == '\n')) {
                // Fine, we support Arm VFPv4 !
                have_vfpv4 = true;
            }

        } while (0);

        if (have_neon)               { features |= SkCpu::NEON    ; }
        if (have_neon && have_vfpv4) { features |= SkCpu::NEON_FMA; }
        if (have_vfpv4)              { features |= SkCpu::VFP_FP16; }
        return features;
    }

#elif defined(SK_CPU_ARM64)         && \
      defined(SK_BUILD_FOR_ANDROID) && \
     !defined(SK_BUILD_FOR_ANDROID_FRAMEWORK)
    #include "cpu-features.h"

    static uint32_t read_cpu_features() {
        uint32_t features = 0;

        uint64_t android_features = android_getCpuFeatures();
        if (android_features & ANDROID_CPU_ARM64_FEATURE_CRC32) { features |= SkCpu::CRC32; }
        return features;
    }

#else
    static uint32_t read_cpu_features() {
        return 0;
    }

#endif

uint32_t SkCpu::gCachedFeatures = 0;

void SkCpu::CacheRuntimeFeatures() {
    static SkOnce once;
    once([] { gCachedFeatures = read_cpu_features(); });
}
